@@ -1,7 +1,7 @@
 ;;; dgnushack.el --- a hack to set the load path for byte-compiling
-;; Copyright (C) 1994,95,96,97,98 Free Software Foundation, Inc.
+;; Copyright (C) 1994,95,96,97,98,99 Free Software Foundation, Inc.
 
-;; Author: Lars Magne Ingebrigtsen <larsi@ifi.uio.no>
+;; Author: Lars Magne Ingebrigtsen <larsi@gnus.org>
 ;; Version: 4.19
 ;; Keywords: news, path
 
@@ -29,10 +29,32 @@
 (fset 'facep 'ignore)
 
 (require 'cl)
+
+;; If we are building w3 in a different directory than the source
+;; directory, we must read *.el from source directory and write *.elc
+;; into the building directory.  For that, we define this function
+;; before loading bytecomp.  Bytecomp doesn't overwrite this function.
+(defun byte-compile-dest-file (filename)
+  "Convert an Emacs Lisp source file name to a compiled file name.
+ In addition, remove directory name part from FILENAME."
+  (setq filename (byte-compiler-base-file-name filename))
+  (setq filename (file-name-sans-versions filename))
+  (setq filename (file-name-nondirectory filename))
+  (if (memq system-type '(win32 w32 mswindows windows-nt))
+      (setq filename (downcase filename)))
+  (cond ((eq system-type 'vax-vms)
+        (concat (substring filename 0 (string-match ";" filename)) "c"))
+       ((string-match emacs-lisp-file-regexp filename)
+        (concat (substring filename 0 (match-beginning 0)) ".elc"))
+       (t (concat filename ".elc"))))
+
 (require 'bytecomp)
-(push "~/lisp/custom" load-path)
-(push "." load-path)
-(load "./lpath.el")
+
+(defvar srcdir (or (getenv "srcdir") "."))
+
+(push srcdir load-path)
+;(push "/usr/share/emacs/site-lisp" load-path)
+(load (expand-file-name "lpath.el" srcdir) nil t)
 
 (defalias 'device-sound-enabled-p 'ignore)
 (defalias 'play-sound-file 'ignore)
@@ -48,11 +70,11 @@
     (fset 'x-defined-colors 'ignore)
     (fset 'read-color 'ignore)))
 
-(setq byte-compile-warnings
-      '(free-vars unresolved callargs redefine))
-
-(defun dgnushack-compile ()
+(defun dgnushack-compile (&optional warn)
   ;;(setq byte-compile-dynamic t)
+  (unless warn
+    (setq byte-compile-warnings
+	  '(free-vars unresolved callargs redefine)))
   (unless (locate-library "cus-edit")
     (error "You do not seem to have Custom installed.
 Fetch it from <URL:http://www.dina.kvl.dk/~abraham/custom/>.
@@ -61,19 +83,24 @@ You also then need to add the following to the lisp/dgnushack.el file:
      (push \"~/lisp/custom\" load-path)
 
 Modify to suit your needs."))
-  (let ((files (directory-files "." nil "^[^=].*\\.el$"))
+  (let ((files (directory-files srcdir nil "^[^=].*\\.el$"))
 	(xemacs (string-match "XEmacs" emacs-version))
 	;;(byte-compile-generate-call-tree t)
 	file elc)
     (condition-case ()
-	(require 'w3-forms)
-      (error (setq files (delete "nnweb.el" (delete "nnlistserv.el" files)))))
+ 	(require 'w3-forms)
+      (error
+       (dolist (file '("nnweb.el" "nnlistserv.el" "nnultimate.el"
+		       "nnslashdot.el" "nnwarchive.el" "webmail.el"))
+	 (setq files (delete file files)))))
     (while (setq file (pop files))
+      (setq file (expand-file-name file srcdir))
       (when (or (and (not xemacs)
-		     (not (member file '("gnus-xmas.el" "gnus-picon.el"
-					 "messagexmas.el" "nnheaderxm.el"
-					 "smiley.el" "x-overlay.el"))))
-		(and xemacs
+ 		     (not (member (file-name-nondirectory file)
+ 				  '("gnus-xmas.el" "gnus-picon.el"
+ 				    "messagexmas.el" "nnheaderxm.el"
+ 				    "smiley.el" "x-overlay.el"))))
+ 		(and xemacs
 		     (not (member file '("md5.el")))))
 	(when (or (not (file-exists-p (setq elc (concat file "c"))))
 		  (file-newer-than-file-p file elc))
